@@ -1,5 +1,5 @@
 /* scanners.c -- file & directory name manipulations
-   Copyright (C) 1986, 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1986, 1995, 1996, 1999, 2000 Free Software Foundation, Inc.
    Written by Greg McGary <gkm@gnu.ai.mit.edu>
 
    This program is free software; you can redistribute it and/or modify
@@ -19,54 +19,61 @@
 #include <config.h>
 #include <stdio.h>
 #include <ctype.h>
-#if HAVE_FCNTL_H
-# include <fcntl.h>
-#endif
+#include <fcntl.h>
 #include <getopt.h>
-#include "xstdlib.h"
-#include "xstddef.h"
-#include "xunistd.h"
-#include "xsysstat.h"
-#include "xstring.h"
-#include "xmalloc.h"
+#include <stdlib.h>
+#include <stddef.h>
+#include <unistd.h>
+#include <sys/stat.h>
+#include <string.h>
+#include <errno.h>
+#include "xalloc.h"
 #include "xnls.h"
 #include "error.h"
 #include "scanners.h"
 #include "tokflags.h"
-
-#define DEBUG(args) /* printf args */
+#include "iduglobal.h"
 
 struct obstack lang_args_obstack;
 struct lang_args *lang_args_default = 0;
 struct lang_args *lang_args_list = 0;
 struct obstack tokens_obstack;
+int log_8_member_files = 0;
 
-extern void usage __P((void));
+extern void usage (void);
 extern char *program_name;
 
 /****************************************************************************/
 
-struct lang_args **parse_language_map_file __P((char const *file_name, struct lang_args **next_ptr));
-char *read_language_map_file __P((char const *file_name));
-void tokenize_args_string __P((char *args_string, int *argcp, char ***argvp));
+struct lang_args **parse_language_map_file (char const *file_name, struct lang_args **next_ptr);
+char *read_language_map_file (char const *file_name);
+void tokenize_args_string (char *args_string, int *argcp, char ***argvp);
 
-static struct token *get_token_c __P((FILE *in_FILE, void const *args, int *flags));
-static void *parse_args_c __P((char **argv, int argc));
-static void help_me_c __P((void));
+static struct token *get_token_c (FILE *in_FILE, void const *args, int *flags);
+static void *parse_args_c (char **argv, int argc);
+static void help_me_c (void);
+static void help_me_cpp (void);
 
-static struct token *get_token_asm __P((FILE *in_FILE, void const *args, int *flags));
-static void *parse_args_asm __P((char **argv, int argc));
-static void help_me_asm __P((void));
+static struct token *get_token_asm (FILE *in_FILE, void const *args, int *flags);
+static void *parse_args_asm (char **argv, int argc);
+static void help_me_asm (void);
+
+static struct token *get_token_text (FILE *in_FILE, void const *args, int *flags);
+static void *parse_args_text (char **argv, int argc);
+static void help_me_text (void);
+
+static struct token *get_token_perl (FILE *in_FILE, void const *args, int *flags);
+static void *parse_args_perl (char **argv, int argc);
+static void help_me_perl (void);
 
-static struct token *get_token_text __P((FILE *in_FILE, void const *args, int *flags));
-static void *parse_args_text __P((char **argv, int argc));
-static void help_me_text __P((void));
 
 struct language languages_0[] =
 {
   { "C", parse_args_c, get_token_c, help_me_c },
+  { "C++", parse_args_c, get_token_c, help_me_cpp },
   { "asm", parse_args_asm, get_token_asm, help_me_asm },
   { "text", parse_args_text, get_token_text, help_me_text },
+  { "perl", parse_args_perl, get_token_perl, help_me_perl }
 };
 struct language const *languages_N = &languages_0[cardinalityof (languages_0)];
 
@@ -85,7 +92,7 @@ void
 language_save_arg (char *arg)
 {
   static char horizontal_space[] = " \t";
-  char *lang_name = strtok (arg, ":");
+  char *lang_name = strsep (&arg, ":");
   struct language *lang = get_language (lang_name);
 
   if (lang == 0)
@@ -95,7 +102,7 @@ language_save_arg (char *arg)
     }
   if (lang->lg_argc == 0)
     lang->lg_argv[lang->lg_argc++] = program_name;
-  lang->lg_argv[lang->lg_argc++] = strtok (0, horizontal_space);
+  lang->lg_argv[lang->lg_argc++] = strsep (&arg, horizontal_space);
 }
 
 void
@@ -188,7 +195,7 @@ parse_language_map_file (char const *file_name, struct lang_args **next_ptr)
 	  continue;
 	}
 
-      new_args = OBSTACK_ALLOC (&lang_args_obstack, struct lang_args, 1);
+      new_args = obstack_alloc (&lang_args_obstack, sizeof(struct lang_args) * 1);
       if (new_args == 0)
 	error (1, 0, _("can't allocate language args: memory exhausted"));
       new_args->la_pattern = obstack_copy0 (&lang_args_obstack, lmp, pattern_size);
@@ -257,7 +264,7 @@ read_language_map_file (char const *file_name)
   if (fstat (map_fd, &st) < 0)
     error (1, errno, _("can't get size of map file `%s'"), file_name);
 
-  lang_map_buffer = MALLOC (char, st.st_size + 2);
+  lang_map_buffer = xmalloc (sizeof(char) * (st.st_size + 2));
   if (lang_map_buffer == 0)
     error (1, 0, _("can't allocate language args: memory exhausted"));
   lang_map_buffer[st.st_size] = '\n';
@@ -280,55 +287,59 @@ void
 tokenize_args_string (char *args_string, int *argcp, char ***argvp)
 {
   static char horizontal_space[] = " \t";
-  char **argv_0 = MALLOC (char *, strlen (args_string) / 2);
+  char **argv_0 = xmalloc (sizeof(char *) * strlen (args_string) / 2);
   char **argv = argv_0;
   char *arg;
 
   *argv++ = program_name;
-  arg = strtok (args_string, horizontal_space);
+  arg = strsep (&args_string, horizontal_space);
   while (arg)
     {
       *argv++ = arg;
-      arg = strtok (0, horizontal_space);
+      arg = strsep (&args_string, horizontal_space);
     }
   *argcp = argv - argv_0;
-  *argvp = REALLOC (argv_0, char *, *argcp);
+  *argvp = xrealloc (argv_0, sizeof(char *) * (*argcp));
 }
 
 static void
 set_ushort_ctype (unsigned short *ctype, char const *chars, int type)
 {
   unsigned short *rct = &ctype[1];
+  unsigned char const *uc = (unsigned char const *) chars;
 
-  while (*chars)
-    rct[*chars++] |= type;
+  while (*uc)
+    rct[*uc++] |= type;
 }
 
 static void
 clear_ushort_ctype (unsigned short *ctype, char const *chars, int type)
 {
   unsigned short *rct = &ctype[1];
+  unsigned char const *uc = (unsigned char const *) chars;
 
-  while (*chars)
-    rct[*chars++] &= ~type;
+  while (*uc)
+    rct[*uc++] &= ~type;
 }
 
 static void
 set_uchar_ctype (unsigned char *ctype, char const *chars, int type)
 {
   unsigned char *rct = &ctype[1];
+  unsigned char const *uc = (unsigned char const *) chars;
 
-  while (*chars)
-    rct[*chars++] |= type;
+  while (*uc)
+    rct[*uc++] |= type;
 }
 
 static void
 clear_uchar_ctype (unsigned char *ctype, char const *chars, int type)
 {
   unsigned char *rct = &ctype[1];
+  unsigned char const *uc = (unsigned char const *) chars;
 
-  while (*chars)
-    rct[*chars++] &= ~type;
+  while (*uc)
+    rct[*uc++] &= ~type;
 }
 
 /*************** C & C++ ****************************************************/
@@ -417,6 +428,17 @@ C language:\n\
 "));
 }
 
+static void
+help_me_cpp (void)
+{
+  printf (_("\
+C++ language:\n\
+  -k,--keep=CHARS        Allow CHARS in single-token strings, keep the result\n\
+  -i,--ignore=CHARS      Allow CHARS in single-token strings, toss the result\n\
+  -u,--strip-underscore  Strip a leading underscore from single-token strings\n\
+"));
+}
+
 static void *
 parse_args_c (char **argv, int argc)
 {
@@ -432,7 +454,7 @@ parse_args_c (char **argv, int argc)
     {
       tmp_string = strdup (*argv);
       tokenize_args_string (tmp_string, &argc, &argv);
-      args = MALLOC (struct args_c, 1);
+      args = xmalloc (sizeof(struct args_c) * 1);
       args->strip_underscore = 0;
       args->ctype = ctype_c;
     }
@@ -472,6 +494,94 @@ parse_args_c (char **argv, int argc)
   return args;
 }
 
+unsigned char *scanner_buffer;
+
+#define SCAN_CPP_DIRECTIVE						\
+  do									\
+    {									\
+      new_line = 0;							\
+      /* Cope with leading whitespace before CPP lines */		\
+      while (c == ' ' || c == '\t')					\
+	c = getc (in_FILE);						\
+      if (c == '\n')							\
+	{								\
+	  new_line = 1;							\
+	  goto top;							\
+	}								\
+      else if (c != '#')						\
+	goto next;							\
+      c = getc (in_FILE);						\
+      while (ISBORING (c))						\
+	c = getc (in_FILE);						\
+      if (!ISID1ST (c))							\
+	goto next;							\
+      id = scanner_buffer;						\
+      *id++ = c;							\
+      while (ISIDREST (c = getc (in_FILE)))				\
+	*id++ = c;							\
+      *id = '\0';							\
+      if (strequ (scanner_buffer, "include"))				\
+	{								\
+	  while (c == ' ' || c == '\t')					\
+	    c = getc (in_FILE);						\
+	  if (c == '\n')						\
+	    {								\
+	      new_line = 1;						\
+	      goto top;							\
+	    }								\
+	  id = scanner_buffer;						\
+	  if (c == '"')							\
+	    {								\
+	      c = getc (in_FILE);					\
+	      while (c != '\n' && c != '"' && c != EOF)			\
+		{							\
+		  *id++ = c;						\
+		  c = getc (in_FILE);					\
+		}							\
+	      *flags = TOK_STRING;					\
+	    }								\
+	  else if (c == '<')						\
+	    {								\
+	      c = getc (in_FILE);					\
+	      while (c != '\n' && c != '>' && c != EOF)			\
+		{							\
+		  *id++ = c;						\
+		  c = getc (in_FILE);					\
+		}							\
+	      *flags = TOK_STRING;					\
+	    }								\
+	  else if (ISID1ST (c))						\
+	    {								\
+	      *id++ = c;						\
+	      while (ISIDREST (c = getc (in_FILE)))			\
+		*id++ = c;						\
+	      *flags = TOK_NAME;					\
+	    }								\
+	  else								\
+	    {								\
+	      while (c != '\n' && c != EOF)				\
+		c = getc (in_FILE);					\
+	      new_line = 1;						\
+	      goto top;							\
+	    }								\
+	  while (c != '\n' && c != EOF)					\
+	    c = getc (in_FILE);						\
+	  new_line = 1;							\
+	  obstack_grow0 (&tokens_obstack, scanner_buffer,		\
+			 id - scanner_buffer);				\
+	  return (struct token *) obstack_finish (&tokens_obstack);	\
+	}								\
+      if (strnequ (scanner_buffer, "if", 2)				\
+	  || strequ (scanner_buffer, "define")				\
+	  || strequ (scanner_buffer, "elif")	/* ansi C */		\
+	  || strequ (scanner_buffer, "undef"))				\
+	goto next;							\
+      while ((c != '\n') && (c != EOF))					\
+	c = getc (in_FILE);						\
+      new_line = 1;							\
+      goto top;								\
+  } while (0)
+
 
 /* Grab the next identifier from the C source file.  This state
    machine is built for speed, not elegance.  */
@@ -482,89 +592,15 @@ get_token_c (FILE *in_FILE, void const *args, int *flags)
 #define ARGS ((struct args_c const *) args)
   static int new_line = 1;
   unsigned short const *rct = &ARGS->ctype[1];
-  char id_0[BUFSIZ];
-  char *id = id_0;
-  int c;
+  unsigned char *id = scanner_buffer;
+  int c; int d;
 
-  obstack_blank (&tokens_obstack, offsetof (struct token, tok_name));
+  obstack_blank (&tokens_obstack, OFFSETOF_TOKEN_NAME);
 
 top:
   c = getc (in_FILE);
   if (new_line)
-    {
-      new_line = 0;
-      if (c != '#')
-	goto next;
-      c = getc (in_FILE);
-      while (ISBORING (c))
-	c = getc (in_FILE);
-      if (!ISID1ST (c))
-	goto next;
-      id = id_0;
-      *id++ = c;
-      while (ISIDREST (c = getc (in_FILE)))
-	*id++ = c;
-      *id = '\0';
-      if (strequ (id_0, "include"))
-	{
-	  while (c == ' ' || c == '\t')
-	    c = getc (in_FILE);
-	  if (c == '\n')
-	    {
-	      new_line = 1;
-	      goto top;
-	    }
-	  id = id_0;
-	  if (c == '"')
-	    {
-	      c = getc (in_FILE);
-	      while (c != '\n' && c != EOF && c != '"')
-		{
-		  *id++ = c;
-		  c = getc (in_FILE);
-		}
-	      *flags = TOK_STRING;
-	    }
-	  else if (c == '<')
-	    {
-	      c = getc (in_FILE);
-	      while (c != '\n' && c != EOF && c != '>')
-		{
-		  *id++ = c;
-		  c = getc (in_FILE);
-		}
-	      *flags = TOK_STRING;
-	    }
-	  else if (ISID1ST (c))
-	    {
-	      *id++ = c;
-	      while (ISIDREST (c = getc (in_FILE)))
-		*id++ = c;
-	      *flags = TOK_NAME;
-	    }
-	  else
-	    {
-	      while (c != '\n' && c != EOF)
-		c = getc (in_FILE);
-	      new_line = 1;
-	      goto top;
-	    }
-	  while (c != '\n' && c != EOF)
-	    c = getc (in_FILE);
-	  new_line = 1;
-	  obstack_grow0 (&tokens_obstack, id_0, id - id_0);
-	  return (struct token *) obstack_finish (&tokens_obstack);
-	}
-      if (strnequ (id_0, "if", 2)
-	  || strequ (id_0, "define")
-	  || strequ (id_0, "elif")	/* ansi C */
-	  || strequ (id_0, "undef"))
-	goto next;
-      while ((c != '\n') && (c != EOF))
-	c = getc (in_FILE);
-      new_line = 1;
-      goto top;
-    }
+    SCAN_CPP_DIRECTIVE;
 
 next:
   while (ISBORING (c))
@@ -573,7 +609,7 @@ next:
   switch (c)
     {
     case '"':
-      id = id_0;
+      id = scanner_buffer;
       *id++ = c = getc (in_FILE);
       for (;;)
 	{
@@ -589,19 +625,19 @@ next:
 	  break;
 	}
       *--id = '\0';
-      id = id_0;
-      while (ISSTRKEEP (*id))
+      id = scanner_buffer;
+      while (ISSTRKEEP (d = *id))
 	id++;
-      if (*id || id == id_0)
+      if (*id || id == scanner_buffer)
 	{
 	  c = getc (in_FILE);
 	  goto next;
 	}
       *flags = TOK_STRING;
-      if (ARGS->strip_underscore && id_0[0] == '_' && id_0[1])
-	obstack_grow0 (&tokens_obstack, id_0 + 1, id - id_0 - 1);
+      if (ARGS->strip_underscore && scanner_buffer[0] == '_' && scanner_buffer[1])
+	obstack_grow0 (&tokens_obstack, scanner_buffer + 1, id - scanner_buffer - 1);
       else
-	obstack_grow0 (&tokens_obstack, id_0, id - id_0);
+	obstack_grow0 (&tokens_obstack, scanner_buffer, id - scanner_buffer);
       return (struct token *) obstack_finish (&tokens_obstack);
 
     case '\'':
@@ -661,7 +697,7 @@ next:
 	  obstack_free (&tokens_obstack, obstack_finish (&tokens_obstack));
 	  return 0;
 	}
-      id = id_0;
+      id = scanner_buffer;
       *id++ = c;
       if (ISID1ST (c))
 	{
@@ -684,7 +720,7 @@ next:
 	}
       ungetc (c, in_FILE);
       *flags |= TOK_LITERAL;
-      obstack_grow0 (&tokens_obstack, id_0, id - id_0);
+      obstack_grow0 (&tokens_obstack, scanner_buffer, id - scanner_buffer);
       return (struct token *) obstack_finish (&tokens_obstack);
     }
 #undef ARGS
@@ -777,7 +813,7 @@ static struct option const long_options_asm[] =
   { "keep", required_argument, 0, 'k' },
   { "ignore", required_argument, 0, 'i' },
   { "strip-underscore", no_argument, 0, 'u' },
-  { "no-cpp", no_argument, 0, 'p' },
+  { "no-cpp", no_argument, 0, 'n' },
   { 0 }
 };
 
@@ -809,7 +845,7 @@ parse_args_asm (char **argv, int argc)
     {
       tmp_string = strdup (*argv);
       tokenize_args_string (tmp_string, &argc, &argv);
-      args = MALLOC (struct args_asm, 1);
+      args = xmalloc (sizeof(struct args_asm) * 1);
       args->strip_underscore = 0;
       args->ctype = ctype_asm;
       args->handle_cpp = 1;
@@ -868,49 +904,15 @@ get_token_asm (FILE *in_FILE, void const *args, int *flags)
 #define ARGS ((struct args_asm const *) args)
   static int new_line = 1;
   unsigned char const *rct = &ARGS->ctype[1];
-  char id_0[BUFSIZ];
-  char *id = id_0;
-  int c;
+  unsigned char *id = scanner_buffer;
+  int c, d;
 
-  obstack_blank (&tokens_obstack, offsetof (struct token, tok_name));
+  obstack_blank (&tokens_obstack, OFFSETOF_TOKEN_NAME);
 
 top:
   c = getc (in_FILE);
   if (ARGS->handle_cpp > 0 && new_line)
-    {
-      new_line = 0;
-      if (c != '#')
-	goto next;
-      while (ISBORING (c))
-	c = getc (in_FILE);
-      if (!ISID1ST (c))
-	goto next;
-      id = id_0;
-      *id++ = c;
-      while (ISIDREST (c = getc (in_FILE)))
-	*id++ = c;
-      *id = '\0';
-      if (strequ (id_0, "include"))
-	{
-	  while (c != '"' && c != '<')
-	    c = getc (in_FILE);
-	  id = id_0;
-	  *id++ = c = getc (in_FILE);
-	  while ((c = getc (in_FILE)) != '"' && c != '>')
-	    *id++ = c;
-	  *flags = TOK_STRING;
-	  obstack_grow0 (&tokens_obstack, id_0, id - id_0);
-	  return (struct token *) obstack_finish (&tokens_obstack);
-	}
-      if (strnequ (id_0, "if", 2)
-	  || strequ (id_0, "define")
-	  || strequ (id_0, "undef"))
-	goto next;
-      while (c != '\n')
-	c = getc (in_FILE);
-      new_line = 1;
-      goto top;
-    }
+    SCAN_CPP_DIRECTIVE;
 
 next:
   while (ISBORING (c))
@@ -961,7 +963,7 @@ next:
       goto next;
     }
 
-  id = id_0;
+  id = scanner_buffer;
   if (ARGS->strip_underscore && c == '_' && !ISID1ST (c = getc (in_FILE)))
     {
       obstack_grow0 (&tokens_obstack, "_", 1);
@@ -990,12 +992,12 @@ next:
     }
 
   *id = '\0';
-  for (id = id_0; *id; id++)
-    if (ISIGNORE (*id))
+  for (id = scanner_buffer; *id; id++)
+    if (ISIGNORE (d = *id))
       goto next;
   ungetc (c, in_FILE);
   *flags |= TOK_LITERAL;
-  obstack_grow0 (&tokens_obstack, id_0, id - id_0);
+  obstack_grow0 (&tokens_obstack, scanner_buffer, id - scanner_buffer);
   return (struct token *) obstack_finish (&tokens_obstack);
 #undef ARGS
 }
@@ -1111,7 +1113,7 @@ parse_args_text (char **argv, int argc)
     {
       tmp_string = strdup (*argv);
       tokenize_args_string (tmp_string, &argc, &argv);
-      args = MALLOC (struct args_text, 1);
+      args = xmalloc (sizeof(struct args_text) * 1);
       args->ctype = ctype_text;
     }
 
@@ -1122,7 +1124,7 @@ parse_args_text (char **argv, int argc)
 			      long_options_text, (int *) 0);
       if (optc < 0)
 	break;
-      if ((optc == 'k' || optc == 'i') && args->ctype == ctype_text)
+      if ((optc == 'x' || optc == 'i') && args->ctype == ctype_text)
 	args->ctype = CLONE (ctype_text, unsigned char, cardinalityof (ctype_text));
       switch (optc)
 	{
@@ -1153,12 +1155,11 @@ static struct token *
 get_token_text (FILE *in_FILE, void const *args, int *flags)
 {
 #define ARGS ((struct args_text const *) args)
-  static char id_0[BUFSIZ];
   unsigned char const *rct = &ARGS->ctype[1];
   int c;
-  char *id = id_0;
+  unsigned char *id = scanner_buffer;
 
-  obstack_blank (&tokens_obstack, offsetof (struct token, tok_name));
+  obstack_blank (&tokens_obstack, OFFSETOF_TOKEN_NAME);
 
 top:
   c = getc (in_FILE);
@@ -1169,7 +1170,7 @@ top:
       obstack_free (&tokens_obstack, obstack_finish (&tokens_obstack));
       return 0;
     }
-  id = id_0;
+  id = scanner_buffer;
   *id++ = c;
   if (ISID1ST (c))
     {
@@ -1195,7 +1196,7 @@ top:
 
   ungetc (c, in_FILE);
   *flags |= TOK_LITERAL;
-  obstack_grow0 (&tokens_obstack, id_0, id - id_0);
+  obstack_grow0 (&tokens_obstack, scanner_buffer, id - scanner_buffer);
   return (struct token *) obstack_finish (&tokens_obstack);
 #undef ARGS
 }
@@ -1210,3 +1211,298 @@ top:
 #undef ISEOF
 #undef ISBORING
 #undef ISIDSQUEEZE
+
+/*************** Perl *******************************************************/
+
+#define I1	0x01		/* 1st char of an identifier [a-zA-Z_] */
+#define NM	0x02		/* digit [0-9a-fA-FxX] */
+#define SQ	0x04		/* squeeze these out (.,',-) */
+#define CM	0x08		/* Comment Character '#' */
+#define EQ	0x10		/* Start of Documentation Character '=' */
+#define Q1	0x20		/* Start of Single Quote ''' */
+#define Q2	0x40		/* Start of Double Quote '"' */
+#define EF	0x80		/* EOF */
+#define NL	0x100		/* '\n' character */
+#define ES	0x200		/* Escape '\\' character */
+
+/* Text character classes */
+#define ISID1ST(c)	((rct)[c] & (I1))
+#define ISIDREST(c)	((rct)[c] & (I1|NM|SQ|Q1))
+#define ISNUMBER(c)	((rct)[c] & (NM))
+#define ISIDSQUEEZE(c)	((rct)[c] & (SQ))
+#define ISCOMMENT(c)	((rct)[c] & (CM))
+#define ISEQUALS(c)	((rct)[c] & (EQ))
+#define ISQUOTE1(c)	((rct)[c] & (Q1))
+#define ISQUOTE2(c)	((rct)[c] & (Q2))
+#define ISNEWLINE(c)	((rct)[c] & (NL))
+#define ISEOF(c)	((rct)[c] & (EF))
+#define ISBORING(c)	(!((rct)[c] & (I1|NL|NM|CM|EQ|Q1|Q2|ES|EF)))
+
+static unsigned short ctype_perl[257] =
+{
+  EF,
+/*      0       1       2       3       4       5       6       7   */
+/*    -----   -----   -----   -----   -----   -----   -----   ----- */
+/*000*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*010*/ 0,	0,	NL,	0,	0,	0,	0,	0,
+/*020*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*030*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*040*/ 0,	0,	Q2,	CM,	0,	0,	0,	Q1,
+/*050*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*060*/ NM,	NM,	NM,	NM,	NM,	NM,	NM,	NM,
+/*070*/ NM,	NM,	0,	0,	0,	EQ,	0,	0,
+/*100*/ 0,	I1|NM,	I1|NM,	I1|NM,	I1|NM,	I1|NM,	I1|NM,	I1,
+/*110*/ I1,	I1,	I1,	I1,	I1|NM,	I1,	I1,	I1,
+/*120*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	I1,
+/*130*/ I1|NM,	I1,	I1,	0,	ES,	0,	0,	I1,
+/*140*/ 0,	I1|NM,	I1|NM,	I1|NM,	I1|NM,	I1|NM,	I1|NM,	I1,
+/*150*/ I1,	I1,	I1,	I1,	I1|NM,	I1,	I1,	I1,
+/*160*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	I1,
+/*170*/ I1|NM,	I1,	I1,	0,	0,	0,	0,	0,
+/*200*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*210*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*220*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*230*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*240*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*250*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*260*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*270*/ 0,	0,	0,	0,	0,	0,	0,	0,
+/*300*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	I1,
+/*310*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	I1,
+/*320*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	0,
+/*330*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	I1,
+/*340*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	I1,
+/*350*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	I1,
+/*360*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	0,
+/*370*/ I1,	I1,	I1,	I1,	I1,	I1,	I1,	I1,
+};
+
+struct args_perl
+{
+  int exclude_dtags;
+  unsigned short *ctype;
+};
+
+static struct args_perl args_perl = { 1, ctype_perl };
+
+static struct option const long_options_perl[] =
+{
+  { "include", required_argument, 0, 'i' },
+  { "exclude", required_argument, 0, 'x' },
+  { "dtags", no_argument, 0, 'd' },
+  { 0 }
+};
+
+static void
+help_me_perl (void)
+{
+  printf (_("\
+Perl language:\n\
+  -i,--include=CHAR-CLASS  Treat characters of CHAR-CLASS as token constituents\n\
+  -x,--exclude=CHAR-CLASS  Treat characters of CHAR-CLASS as token delimiters\n\
+  -d,--dtags  Include documentation tags\n\
+"));
+}
+
+static void *
+parse_args_perl (char **argv, int argc)
+{
+  char *tmp_string = 0;
+  struct args_perl *args;
+
+  if (argv == 0 || *argv == 0)
+    return &args_perl;
+
+  if (argc)
+    args = &args_perl;
+  else
+    {
+      tmp_string = strdup (*argv);
+      tokenize_args_string (tmp_string, &argc, &argv);
+      args = xmalloc (sizeof(struct args_perl));
+      args->exclude_dtags = 1;
+      args->ctype = ctype_perl;
+    }
+
+  optind = 0;
+  for (;;)
+    {
+      int optc = getopt_long (argc, argv, "i:x:d",
+			      long_options_perl, (int *) 0);
+      if (optc < 0)
+	break;
+      if ((optc == 'x' || optc == 'i') && args->ctype == ctype_perl)
+	args->ctype = CLONE (ctype_perl, unsigned short, cardinalityof (ctype_perl));
+      switch (optc)
+	{
+	case 'd':
+	  args->exclude_dtags = 0;
+	  break;
+
+	case 'i':
+	  set_ushort_ctype (args->ctype, optarg, I1);
+	  break;
+
+	case 'x':
+	  clear_ushort_ctype (args->ctype, optarg, I1);
+	  break;
+
+	default:
+	  usage ();
+	}
+    }
+  if (tmp_string)
+    {
+      free (argv);
+      free (tmp_string);
+    }
+  return args;
+}
+
+/* Grab the next identifier the text source file.  This state machine
+   is built for speed, not elegance.  */
+
+static struct token *
+get_token_perl (FILE *in_FILE, void const *args, int *flags)
+{
+#define ARGS ((struct args_perl const *) args)
+  static int new_line = 1;
+  /*  static char id_0[BUFSIZ]; */
+  unsigned short const *rct = &ARGS->ctype[1];
+  int c, state = 0, skip_doc = 0;
+  /* int comment = 0, d_quote = 0, s_quote = 0, equals = 0; */
+  unsigned char *id = scanner_buffer;
+
+  obstack_blank (&tokens_obstack, OFFSETOF_TOKEN_NAME);
+
+top: 
+  c = getc (in_FILE);
+  while (ISBORING (c))
+    c = getc (in_FILE);
+
+  switch (c)
+    {
+      case '\n':
+        state &= ~CM;		/* comment = 0; */
+        new_line = 1;
+        goto top;
+        break;
+
+      case '\\':
+        c = getc (in_FILE); /* Skip next character */
+        new_line = 0;
+        goto top;
+        break;
+
+      case '#':
+        state |= CM;		/* comment = 1; */
+        break;
+
+      case '\'':
+	if (!skip_doc)
+           state ^= Q1;		/* s_quote = ((s_quote) ? 0 : 1); */
+        break;
+
+      case '\"':
+	if (!skip_doc)
+          state ^= Q2;		/* d_quote = ((d_quote == 1) ? 0 : 1); */
+        break;
+
+      case '=':
+        if (new_line && ARGS->exclude_dtags)
+          {
+            skip_doc = 1;
+            state &= ~EQ;		/* documenation = 0; */
+          }
+        break;
+
+      default: 
+        if (ISEOF (c))
+          {
+            new_line = 1;
+            obstack_free (&tokens_obstack, obstack_finish (&tokens_obstack));
+            return 0;
+          }
+      break;
+    }
+
+  new_line = 0;
+  if (ISCOMMENT(c) || ISQUOTE1(c) || ISQUOTE2(c) ||
+                ISNEWLINE(c) || ISEQUALS(c) || state)
+     goto top;
+
+  id = scanner_buffer;
+  *id++ = c;
+  if (ISID1ST (c))
+    {
+      *flags = TOK_NAME;
+      while (ISIDREST (c = getc (in_FILE)))
+        if (!ISIDSQUEEZE (c))
+          *id++ = c;
+    }
+  else if (ISNUMBER (c))
+    {
+      *flags = TOK_NUMBER;
+      while (ISNUMBER (c = getc (in_FILE)))
+        *id++ = c;
+
+      ungetc (c, in_FILE);
+      goto top;			/* skip all numbers */
+    }
+  else
+    {
+      if (isprint (c))
+        fprintf (stderr, _("junk: `%c'"), c);
+      else
+        fprintf (stderr, _("junk: `\\%03o'"), c);
+      goto top;
+    }
+
+  ungetc (c, in_FILE);
+
+  *id = '\0';
+  if (skip_doc)
+    {
+      if (strequ (scanner_buffer, "cut"))
+        {
+          skip_doc = 0;
+        }
+      else 
+        {
+          state |= EQ;		/* documenation = 1; */
+	}
+      goto top;
+    }
+
+  if (strequ (scanner_buffer, "_"))
+    {
+      goto top;
+    }
+
+  *flags |= TOK_LITERAL;
+  obstack_grow0 (&tokens_obstack, scanner_buffer, id - scanner_buffer);
+  return (struct token *) obstack_finish (&tokens_obstack);
+#undef ARGS
+}
+
+#undef I1
+#undef NM
+#undef SQ
+#undef CM
+#undef EQ
+#undef Q1
+#undef Q2
+#undef EF
+#undef NL
+#undef ES
+#undef ISID1ST
+#undef ISIDREST
+#undef ISNUMBER
+#undef ISIDSQUEEZE
+#undef ISCOMMENT
+#undef ISEQUALS
+#undef ISQUOTE1
+#undef ISQUOTE2
+#undef ISNEWLINE
+#undef ISEOF
+#undef ISBORING
