@@ -1,5 +1,5 @@
 /* walker.c -- nifty file-tree walker
-   Copyright (C) 1986, 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1986, 1995, 1996, 1999, 2000 Free Software Foundation, Inc.
    Written by Greg McGary <gkm@gnu.ai.mit.edu>
 
    This program is free software; you can redistribute it and/or modify
@@ -18,56 +18,64 @@
 */
 
 #include <config.h>
-#include "xsysstat.h"
+#include <sys/stat.h>
 #include <stdio.h>
-#include "xstdlib.h"
-#include "xstddef.h"
-#include "xunistd.h"
-#include "xstring.h"
-#include "xfnmatch.h"
-#include "xdirent.h"
+#include <stdlib.h>
+#include <stddef.h>
+#include <unistd.h>
+#include <string.h>
+#include <fnmatch.h>
+#include <dirent.h>
+#include <errno.h>
 #include "xnls.h"
 #include "idfile.h"
 #include "error.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "dynvec.h"
 #include "scanners.h"
 #include "pathmax.h"
-#include "xalloca.h"
+#include "xalloc.h"
+#include "iduglobal.h"
 
-int walk_dir __P((struct file_link *dir_link));
-struct member_file *get_member_file __P((struct file_link *flink));
-struct lang_args *get_lang_args __P((struct file_link const *flink));
-int walk_sub_dirs __P((struct dynvec *sub_dirs_vec));
-void reparent_children __P((struct file_link *dlink, struct file_link *slink));
-int classify_link __P((struct file_link *flink, struct stat *stp));
-struct file_link *get_link_from_dirent __P((struct dirent *dirent, struct file_link *parent));
-struct file_link *make_link_from_dirent __P((struct dirent *dirent, struct file_link *parent));
-struct file_link *get_link_from_string __P((char const *name, struct file_link *parent));
-struct file_link *make_link_from_string __P((char const *name, struct file_link *parent));
-int lang_wanted __P((char const *lang_name));
-char **append_strings_to_vector __P((char **vector_0, char *string, char *delimiter_class));
-int vector_length __P((char **vector));
-int string_in_vector __P((char const *string, char **vector));
-static int same_as_dot __P((char const *cwd));
-struct file_link const **fill_link_vector __P((struct file_link const **vec_buf, struct file_link const *flink));
-struct file_link const **fill_link_vector_1 __P((struct file_link const **vec_buf, struct file_link const *flink));
-char *fill_dot_dots __P((char *buf, int levels));
-static char *absolute_file_name_1 __P((char *buffer, struct file_link const *flink));
-unsigned long member_file_hash_1 __P((void const *key));
-unsigned long member_file_hash_2 __P((void const *key));
-int member_file_hash_compare __P((void const *x, void const *y));
-unsigned long file_link_hash_1 __P((void const *key));
-unsigned long file_link_hash_2 __P((void const *key));
-int file_link_hash_compare __P((void const *x, void const *y));
-unsigned long dev_ino_hash_1 __P((void const *key));
-unsigned long dev_ino_hash_2 __P((void const *key));
-int dev_ino_hash_compare __P((void const *x, void const *y));
-int symlink_ancestry __P((struct file_link *flink));
+int walker_verbose_flag = 0;
+off_t largest_member_file = 0;
+
+int walk_dir (struct file_link *dir_link);
+struct member_file *get_member_file (struct file_link *flink);
+struct lang_args *get_lang_args (struct file_link const *flink);
+void print_member_file (struct member_file *member);
+int walk_sub_dirs (struct dynvec *sub_dirs_vec);
+void reparent_children (struct file_link *dlink, struct file_link *slink);
+int classify_link (struct file_link *flink, struct stat *stp);
+struct file_link *get_link_from_dirent (struct dirent *dirent, struct file_link *parent);
+struct file_link *make_link_from_dirent (struct dirent *dirent, struct file_link *parent);
+struct file_link *get_link_from_string (char const *name, struct file_link *parent);
+struct file_link *make_link_from_string (char const *name, struct file_link *parent);
+int lang_wanted (char const *lang_name);
+char **append_strings_to_vector (char **vector_0, char *string, char const *delimiter_class);
+int vector_length (char **vector);
+int string_in_vector (char const *string, char **vector);
+static int same_as_dot (char const *cwd);
+struct file_link const **fill_link_vector (struct file_link const **vec_buf, struct file_link const *flink);
+struct file_link const **fill_link_vector_1 (struct file_link const **vec_buf, struct file_link const *flink);
+char *fill_dot_dots (char *buf, int levels);
+static char *absolute_file_name_1 (char *buffer, struct file_link const *flink);
+unsigned long member_file_hash_1 (void const *key);
+unsigned long member_file_hash_2 (void const *key);
+int member_file_hash_compare (void const *x, void const *y);
+unsigned long file_link_hash_1 (void const *key);
+unsigned long file_link_hash_2 (void const *key);
+int file_link_hash_compare (void const *x, void const *y);
+unsigned long dev_ino_hash_1 (void const *key);
+unsigned long dev_ino_hash_2 (void const *key);
+int dev_ino_hash_compare (void const *x, void const *y);
+int symlink_ancestry (struct file_link *flink);
 
 #if HAVE_LINK
-struct file_link *find_alias_link __P((struct file_link *flink, struct stat *stp));
-struct member_file *maybe_get_member_file __P((struct file_link *flink, struct stat *stp));
+
+struct file_link *find_alias_link (struct file_link *flink, struct stat *stp);
+struct member_file *maybe_get_member_file (struct file_link *flink, struct stat *stp);
+
 #endif
 
 #define IS_DOT(s) ((s)[0] == '.' && (s)[1] == '\0')
@@ -77,9 +85,9 @@ struct member_file *maybe_get_member_file __P((struct file_link *flink, struct s
 
 static struct file_link *current_dir_link = 0;
 
-static char white_space[] = " \t\r\n\v\f";
+static char const white_space[] = " \t\r\n\v\f";
 
-char* xgetcwd __P((void));
+char* xgetcwd (void);
 
 
 /****************************************************************************/
@@ -98,7 +106,7 @@ walk_dir (struct file_link *dir_link)
   dirp = opendir (".");
   if (dirp == 0)
     {
-      char *file_name = ALLOCA (char, PATH_MAX);
+      char *file_name = alloca (PATH_MAX);
       absolute_file_name (file_name, dir_link);
       error (0, errno, _("can't read directory `%s' (`.' from `%s')"), file_name, xgetcwd ());
       return 0;
@@ -165,13 +173,12 @@ walk_flink (struct file_link *flink, struct dynvec *sub_dirs_vec)
   if ((old_flags & FL_TYPE_MASK)
       && (old_flags & FL_TYPE_MASK) != (new_flags & FL_TYPE_MASK))
     {
-      char *file_name = ALLOCA (char, PATH_MAX);
+      char *file_name = alloca (PATH_MAX);
       absolute_file_name (file_name, flink);
       error (0, 0, _("notice: `%s' was a %s, but is now a %s!"), file_name,
 	     (FL_IS_FILE (old_flags) ? _("file") : _("directory")),
 	     (FL_IS_FILE (new_flags) ? _("file") : _("directory")));
     }
-  
 
   flink->fl_flags = (old_flags & ~(FL_TYPE_MASK|FL_SYM_LINK)) | new_flags;
   if (FL_IS_DIR (new_flags))
@@ -196,8 +203,13 @@ walk_flink (struct file_link *flink, struct dynvec *sub_dirs_vec)
 #else
       member = get_member_file (flink);
 #endif
-      if (member == 0)
-	return;
+      if (member)
+	{
+	  if (st.st_size > largest_member_file)
+	    largest_member_file = st.st_size;
+	  if (walker_verbose_flag)
+	    print_member_file (member);
+	}
     }
 }
 
@@ -243,7 +255,7 @@ mark_member_file_links (struct idhead *idhp)
 					 0, member_file_qsort_compare);
   struct member_file **end = &members_0[idhp->idh_member_file_table.ht_fill];
   struct member_file **members;
-  int new_index = 0;
+  long new_index = 0;
 
   for (members = members_0; members < end; members++)
     {
@@ -284,8 +296,8 @@ maybe_get_member_file (struct file_link *flink, struct stat *stp)
       int alias_ancestry = symlink_ancestry (alias_link);
       if (member->mf_lang_args != alias_member->mf_lang_args)
 	{
-	  char *file_name = ALLOCA (char, PATH_MAX);
-	  char *alias_file_name = ALLOCA (char, PATH_MAX);
+	  char *file_name = alloca (PATH_MAX);
+	  char *alias_file_name = alloca (PATH_MAX);
 	  absolute_file_name (file_name, flink);
 	  absolute_file_name (alias_file_name, alias_link);
 	  error (0, 0, _("warning: `%s' and `%s' are the same file, but yield different scans!"),
@@ -380,7 +392,7 @@ get_member_file (struct file_link *flink)
 #if 0
       if (member->mf_lang_args != args)
 	{
-	  char *file_name = ALLOCA (char, PATH_MAX);
+	  char *file_name = alloca (PATH_MAX);
 	  absolute_file_name (file_name, flink);
 	  error (0, 0, _("notice: scan parameters changed for `%s'"), file_name);
 	  member->mf_old_index = -1;
@@ -389,7 +401,7 @@ get_member_file (struct file_link *flink)
       member = *slot;
     }
   member->mf_lang_args = args;
-  return *slot;
+  return member;
 }
 
 struct member_file *
@@ -413,12 +425,12 @@ struct lang_args *
 get_lang_args (struct file_link const *flink)
 {
   struct lang_args *args = lang_args_list;
+  char *file_name = alloca (PATH_MAX);
 
   while (args)
     {
       if (strchr (args->la_pattern, SLASH_CHAR))
 	{
-	  char *file_name = ALLOCA (char, PATH_MAX);
 	  absolute_file_name (file_name, flink);
 	  if (fnmatch (args->la_pattern, file_name, MAYBE_FNM_CASEFOLD | FNM_FILE_NAME) == 0)
 	    return (args->la_language ? args : 0);
@@ -434,6 +446,15 @@ get_lang_args (struct file_link const *flink)
 	  ? lang_args_default : 0);
 }
 
+void
+print_member_file (struct member_file *member)
+{
+  char *file_name = alloca (PATH_MAX);
+  absolute_file_name (file_name, member->mf_link);
+  printf ("%ld: %s: %s\n", idh.idh_member_file_table.ht_fill - 1,
+	  member->mf_lang_args->la_language->lg_name, file_name);
+}
+
 
 /****************************************************************************/
 
@@ -468,23 +489,22 @@ exclude_languages (char *lang_names)
 }
 
 char **
-append_strings_to_vector (char **vector_0, char *string, char *delimiter_class)
+append_strings_to_vector (char **vector_0, char *string, char const *delimiter_class)
 {
   char **vector;
   if (vector_0)
     {
       int length = vector_length (vector_0);
-      vector_0 = REALLOC (vector_0, char *,
-			  length + 2 + strlen (string) / 2);
+      vector_0 = xrealloc (vector_0, sizeof(char *) * (length + 2 + strlen (string) / 2));
       vector = &vector_0[length];
     }
   else
-    vector = vector_0 = MALLOC (char *, 2 + strlen (string) / 2);
-  
-  *vector = strtok (string, delimiter_class);
-  while (*vector)
-    *++vector = strtok (0, delimiter_class);
-  return REALLOC (vector_0, char *, ++vector - vector_0);
+    vector = vector_0 = xmalloc (sizeof(char *) * (2 + strlen (string) / 2));
+
+  do
+    *vector = strsep (&string, delimiter_class);
+  while (*vector++);
+  return xrealloc (vector_0, sizeof(char *) * (vector - vector_0));
 }
 
 int
@@ -622,7 +642,7 @@ same_as_dot (char const *cwd)
 int
 chdir_to_link (struct file_link *dir_link)
 {
-  char *to_dir_name = ALLOCA (char, PATH_MAX);
+  char *to_dir_name = alloca (PATH_MAX);
 
   if (current_dir_link == dir_link)
     return 1;
@@ -637,7 +657,7 @@ chdir_to_link (struct file_link *dir_link)
 	error (0, errno, _("can't chdir to `%s'"), to_dir_name);
       else
 	{
-	  char *from_dir_name = ALLOCA (char, PATH_MAX);
+	  char *from_dir_name = alloca (PATH_MAX);
 	  absolute_file_name (from_dir_name, current_dir_link);
 	  error (0, errno, _("can't chdir to `%s' from `%s'"), to_dir_name, from_dir_name);
 	}
@@ -651,15 +671,22 @@ chdir_to_link (struct file_link *dir_link)
 }
 
 char **
-vectorize_string (char *string, char *delimiter_class)
+vectorize_string (char *string, char const *delimiter_class)
 {
-  char **vector_0 = MALLOC (char *, 2 + strlen (string) / 2);
+  char **vector_0 = xmalloc (sizeof(char *) * (2 + strlen (string) / 2));
   char **vector = vector_0;
-  
-  *vector = strtok (string, delimiter_class);
-  while (*vector)
-    *++vector = strtok (0, delimiter_class);
-  return REALLOC (vector_0, char *, ++vector - vector_0);
+
+  *vector++ = strsep (&string, delimiter_class);
+  if (vector[-1])
+    {
+      /* Toss first field if empty */
+      if (vector[-1][0] == '\0')
+	vector--;
+      do
+	*vector = strsep (&string, delimiter_class);
+      while (*vector++);
+    }
+  return xrealloc (vector_0, sizeof(char *) * (vector - vector_0));
 }
 
 void
@@ -710,6 +737,8 @@ classify_link (struct file_link *flink, struct stat *stp)
 #endif
   if (S_ISDIR (stp->st_mode))
     flags |= FL_TYPE_DIR;
+  else if (stp->st_size == 0)
+    return 0;
   else if (S_ISREG (stp->st_mode))
     flags |= FL_TYPE_FILE;
   else
@@ -730,7 +759,8 @@ get_link_from_dirent (struct dirent *dirent, struct file_link *parent)
   new_link = make_link_from_dirent (dirent, parent);
   slot = (struct file_link **) hash_find_slot (&idh.idh_file_link_table, new_link);
   if (HASH_VACANT (*slot))
-    hash_insert_at (&idh.idh_file_link_table, new_link, slot);
+    slot = (struct file_link **) hash_insert_at (&idh.idh_file_link_table,
+						 new_link, slot);
   else
     obstack_free (&idh.idh_file_link_obstack, new_link);
   return *slot;
@@ -745,7 +775,8 @@ get_link_from_string (char const *name, struct file_link *parent)
   new_link = make_link_from_string (name, parent);
   slot = (struct file_link **) hash_find_slot (&idh.idh_file_link_table, new_link);
   if (HASH_VACANT (*slot))
-    hash_insert_at (&idh.idh_file_link_table, new_link, slot);
+    slot = (struct file_link **) hash_insert_at (&idh.idh_file_link_table,
+						 new_link, slot);
   else
     obstack_free (&idh.idh_file_link_obstack, new_link);
   return *slot;
@@ -811,8 +842,8 @@ fill_link_vector_1 (struct file_link const **vec_buf, struct file_link const *fl
 char *
 maybe_relative_file_name (char *buf_0, struct file_link const *to_link, struct file_link const *from_link)
 {
-  struct file_link const **to_link_vec_0 = ALLOCA (struct file_link const *, PATH_MAX/2);
-  struct file_link const **from_link_vec_0 = ALLOCA (struct file_link const *, PATH_MAX/2);
+  struct file_link const **to_link_vec_0 = alloca (sizeof (struct file_link const *) * (PATH_MAX/2));
+  struct file_link const **from_link_vec_0 = alloca (sizeof (struct file_link const *) * (PATH_MAX/2));
   struct file_link const **to_link_vec = to_link_vec_0;
   struct file_link const **from_link_vec = from_link_vec_0;
   struct file_link const **from_link_end;
@@ -825,7 +856,7 @@ maybe_relative_file_name (char *buf_0, struct file_link const *to_link, struct f
     from_link = current_dir_link;
 
   if (alloc_me)
-    buf_0 = MALLOC (char, PATH_MAX);
+    buf_0 = xmalloc (PATH_MAX);
 
   /* Optimize common cases.  */
   if (to_link == from_link)
@@ -884,7 +915,7 @@ maybe_relative_file_name (char *buf_0, struct file_link const *to_link, struct f
     }
 out:
   if (alloc_me)
-    buf_0 = REALLOC (buf_0, char, 1 + strlen (buf_0));
+    buf_0 = xrealloc (buf_0, (1 + strlen (buf_0)));
   return buf_0;
 }
 
@@ -913,7 +944,7 @@ absolute_file_name (char *buf_0, struct file_link const *flink)
   int alloc_me = (buf_0 == 0);
 
   if (alloc_me)
-    buf_0 = MALLOC (char, PATH_MAX);
+    buf_0 = xmalloc (PATH_MAX);
   end = absolute_file_name_1 (buf_0, flink);
   /* Clip the trailing slash.  */
 #if HAVE_LINK
@@ -925,7 +956,7 @@ absolute_file_name (char *buf_0, struct file_link const *flink)
 #endif
   *end++ = '\0';
   if (alloc_me)
-    buf_0 = REALLOC (buf_0, char, end - buf_0);
+    buf_0 = xrealloc (buf_0, (size_t)(end - buf_0));
   return buf_0;
 }
 
